import uvicore
from uvicore.http import Request
from uvicore.typing import Sequence
from fastapi.params import Security
from fastapi.security import SecurityScopes
from uvicore.support.dumper import dump, dd
from uvicore.contracts import UserInfo
from uvicore.http.exceptions import PermissionDenied, NotAuthenticated, HTTPException
from uvicore.http.response import Redirect


@uvicore.service()
class Guard(Security):
    """Uvicore HTTP Route Auth Guard"""

    def __init__(self, scopes: Sequence[str] = []):
        # Ensure scopes is a List, to allow for singles
        if scopes: scopes = [scopes] if isinstance(scopes, str) else list(scopes)

        # Call FastAPI Security Depends with our scope analyzer dependency
        super().__init__(dependency=Scopes(), scopes=scopes, use_cache=True)


class Scopes:

    @property
    def log(self):
        return uvicore.log.name('uvicore.http')

    async def __call__(self, security_scopes: SecurityScopes, request: Request) -> UserInfo:
        # Get user from request.  Will always exist.  If not logged it will be
        # the anonymous user which may still have permissions/scopes to compare
        user: UserInfo = request.user

        # Get scopes List.  These are permissions/scopes defined on this route
        scopes = list(security_scopes.scopes)
        self.log.debug('Auth guard scopes required for {}: {}'.format(request.scope['path'], str(scopes)))

        # If no scopes, allow access
        if not scopes: return user

        # Superadmin always wins, never check scopes
        if user.superadmin: return user

        # User must have ALL scopes defined on the route (an AND statement)
        authorized = True
        missing_scopes = []
        for scope in scopes:
            if scope not in user.permissions:
                authorized = False
                missing_scopes.append(scope)

        # dump('-----------------------------------------------------------------')
        # dump('Auth Route Guard HERE:', scopes, user)
        # dump('-----------------------------------------------------------------')

        # Hack logout of basic auth
        #raise NotAuthenticated(headers={'WWW-Authenticate': 'Basic realm="App1 Web Realm"'})

        # Authorized and authenticated.  Return user in case Guard() is being used
        # as a route parameter using FastAPI Depends.  User will be injected back to param value.
        if authorized:
            self.log.debug(f"Auth guard scopes passed, {user.email} has one of {str(scopes)}")
            return user

        # User does not have the proper scopes, permission denied
        self.log.debug(f"Auth guard scopes FAILED for {user.email}")

        # Not authorized or possibly even authenticated
        # Get unauthenticated handler from auth config
        # Route type (web or api) is added to request.scope from authentication middleware
        route_type = request.scope.get('route_type') or 'api'

        # Get the route auth config based on route_type
        auth_config = uvicore.config.app.auth[route_type]

        # Default unauthenticated is to raise exception.
        # Check for redirect override or additional exception headers
        if auth_config.unauthenticated_handler.redirect:
            location = auth_config.unauthenticated_handler.redirect
            if '/' not in location and '.' in location: location = request.url_for(location)
            separator = '&' if '?' in location else '?'
            referer = separator + "referer=" + str(request.scope.get('path'))
            raise HTTPException(status_code=301, headers={
                'Location': location + referer
            })

        # Check for additional exception headers
        exception_headers = None
        if auth_config.unauthenticated_handler.exception:
            exception_headers = auth_config.unauthenticated_handler.exception.headers

        if user.authenticated:
            # User authenticated but does not have all route scopes
            raise PermissionDenied(missing_scopes, headers=exception_headers)
        else:
            # User is not even logged in
            raise NotAuthenticated(headers=exception_headers)



from uvicore.typing import OrderedDict

# This is the main foundation config.  All items here can be overridden
# when used inside other applications.  Accessible at config('uvicore.foundation')

config = {

    # --------------------------------------------------------------------------
    # Registration Control
    #
    # This lets you control the service provider registrations.  If this app
    # is used as a package inside another app you might not want some things
    # registered in that context.  Use config overrides in your app to change
    # registrations
    # --------------------------------------------------------------------------
    'registers': {
        # 'web_routes': True,
        # 'api_routes': True,
        # 'middleware': True,
        # 'views': True,
        # 'assets': True,
        # 'commands': True,
        # 'models': True,
        # 'tables': True,
        # 'seeders': True,
    },


    # Just experimenting when commands being pulled from a config, probably
    # no need to do this
    # 'commands': {
    #     # Register Ioc commands
    #     'ioc': {
    #         'help': 'Uvicore Ioc (Inversion of Control) Information',
    #         'commands': {
    #             'bindings': 'uvicore.container.commands.ioc.bindings',
    #             'singletons': 'uvicore.container.commands.ioc.singletons',
    #             'overrides': 'uvicore.container.commands.ioc.overrides',
    #             'type': 'uvicore.container.commands.ioc.type',
    #             'get': 'uvicore.container.commands.ioc.get',
    #         },
    #     },

    #     # Register Package commands
    #     'package': {
    #         'help': 'Uvicore Package Information',
    #         'commands': {
    #             'providers': 'uvicore.package.commands.package.providers',
    #             'list': 'uvicore.package.commands.package.list',
    #             'get': 'uvicore.package.commands.package.get',
    #         },
    #     },

    #     # Register Event commands
    #     'event': {
    #         'help': 'Uvicore Event Information',
    #         'commands': {
    #             'list': 'uvicore.events.commands.event.list',
    #             'get': 'uvicore.events.commands.event.get',
    #             'listeners': 'uvicore.events.commands.event.listeners',
    #         },
    #     },
    # },


    # --------------------------------------------------------------------------
    # Package Dependencies (Package Providers)
    #
    # Define all the packages that this package depends on.  At a minimum, only
    # the uvicore.foundation package is required.  The foundation is very minimal
    # and only depends on configuration, logging and the console packages. You
    # must add other core services built into uvicore only if your package
    # requires them.  Services like uvicore.database, uvicore.orm, uvicore.auth
    # uvicore.http, etc...
    # --------------------------------------------------------------------------
    'dependencies': OrderedDict({
        # Careful to only define REQUIRED services for a minimal Uvicore.  Things
        # like auth, db, http, etc... are optional and specified in a packages
        # dependencies if needed.

        # Configuration is used right away in provider register()
        'uvicore.configuration': {
            'provider': 'uvicore.configuration.package.provider.Configuration',
        },

        # I want logging quick as it may be used in providers register() and boot()
        'uvicore.logging': {
            'provider': 'uvicore.logging.package.provider.Logging',
        },

        # Console is required for every uvicore app
        'uvicore.console': {
            'provider': 'uvicore.console.package.provider.Console',
        },

        # Cache is required for every uvicore app
        'uvicore.cache': {
            'provider': 'uvicore.cache.package.provider.Cache',
        },
    }),

}



import uvicore
import traceback
from uvicore.typing import Any, Dict, Optional, List
from starlette.exceptions import HTTPException as _HTTPException
from uvicore.http import status
from uvicore.support.dumper import dump
from uvicore.support.collection import haskey

# This is how you could do it, if you wanted to log
#log = lambda : uvicore.log.name('uvicore.http')

# See https://www.restapitutorial.com/httpstatuscodes.html for a good list to follow

class HTTPException(_HTTPException):
    """Main Base HTTP Exception"""

    # Message is optional and will default the the HTTP status codes TEXT as outlined in the python http module
    # Detail is a more detailed text description of the issue
    # Extra lets you pass in a dict of options or extra information that some handlers may want to use
    def __init__(self,
        status_code: int,
        detail: Optional[str] = None,
        *,
        message: Optional[str] = None,
        exception: Optional[str] = None,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        # Detect if we were raised from another HTTPException with a status_code
        # If so, grab values from that first exception
        if exception is not None and haskey(exception, 'status_code'):
            status_code = exception.status_code
            message = exception.message
            detail = exception.detail
            extra = exception.extra
            headers = exception.headers
            exception = None
        else:
            # Standard try catch that raises HTTPException
            st = traceback.format_exc()
            if 'NoneType' in st: st = None
            message = message or str(exception)
            detail = detail or message
            exception = st

        # Call starlette exception where their detail is my message
        super().__init__(status_code=status_code, detail=message)

        # Swap starlette detail to my message
        self.message = self.detail
        self.detail = detail
        self.exception = exception if uvicore.config.app.debug else None  # Hidden unless in debug mode
        self.extra = extra
        self.headers = headers


class PermissionDenied(HTTPException):
    """Permission Denied Exception"""
    def __init__(self,
        permissions: Optional[List] = None,
        detail: Optional[str] = None,
        *,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        detail = "Permission denied"
        if permissions:
            if type(permissions) != list: permissions = [permissions]
            detail += " to {}".format(permissions)
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            message='Permission Denied',
            detail=detail,
            extra=extra,
            headers=headers,
        )


class NotAuthenticated(HTTPException):
    """Not Authenticated Exception"""
    def __init__(self,
        detail: Optional[str] = None,
        *,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            message='Not Authenticated',
            detail=detail,
            extra=extra,
            headers=headers,
        )


class InvalidCredentials(HTTPException):
    """Invalid Credentials Exception"""
    def __init__(self,
        detail: Optional[str] = None,
        *,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            message='Invalid Credentials',
            detail=detail,
            extra=extra,
            headers=headers,
        )


class NotFound(HTTPException):
    """Not Found Exception"""
    def __init__(self,
        detail: Optional[str] = None,
        *,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            message='Not Found',
            detail=detail,
            extra=extra,
            headers=headers,
        )


class BadParameter(HTTPException):
    """Bad Parameter"""
    def __init__(self,
        detail: Optional[str] = None,
        *,
        exception: Optional[str] = None,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            message='Bad Parameter',
            detail=detail,
            exception=exception,
            extra=extra,
            headers=headers,
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

class Field:
    column: str
    name: Optional[str]
    primary: Optional[bool]
    title: Optional[str]
    description: Optional[str]
    default: Optional[Any]
    sortable: Optional[bool]
    searchable: Optional[bool]
    read_only: Optional[bool]
    write_only: Optional[bool]
    callback: Optional[Any]
    evaluate: Optional[Any]
    relation: Optional[Relation]
    json: Optional[bool]
    properties: Optional[Dict]

    min_length: Optional[int] = None
    max_length: Optional[int] = None
    example: Optional[Any] = None

# At bottom due to circular issues between these two contracts
from .relation import Relation  # isort:skip



import uvicore
from uvicore.orm.fields import Field
from uvicore.support.dumper import dd, dump
from uvicore.support.printer import pretty_call, register_pretty
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic.main import ModelMetaclass as PydanticMetaclass
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, Sequence

import sqlalchemy as sa

# try:
# except ImportError:  # pragma: nocover
#     pass
#     #sa = None  # type: ignore
#     #ClauseElement = None  # type: ignore


# Think of this metaclass as all the STATIC methods similar to @classmethod
# but different in that they become INVISIBLE to the instance of the class
# which is ideal so pydantic doesn't see these and clobber any actual 'fields'
# that have the same name.  This means I can have a field called 'email' and a
# true static class method called `email()` and pydantic won't complain.  I can use
# the static via User.email() and the instatnce from user.email.
# See https://stackoverflow.com/questions/59341761/what-are-the-differences-between-a-classmethod-and-a-metaclass-method
# for a detailed difference between using @classmethod for statics vs a metaclass


@uvicore.service()
class ModelMetaclass(PydanticMetaclass):

    # Testing of duplicate field
    # Only works if EACH model sets a metaclass=ModelMetaclass
    # You cannot set the metaclass at the Module.py level
    # def email(entity):
    #     return 'email on meta'

    # def slug(entity):
    #     return 'slug on meta'


    # Remember all metaclass method ARE @classmethods
    # Remember any method here DO NOT clash with pydantic field names
    # Remember any method here do NOT show up in code intellisense because its a metaclass
    #   If you want it in code intellisense, it must be in model.py and ModelInterface
    #   but then it WILL clash with any pydantic fields of the same name.
    # Basically anything here should be used privately from inside the ORM.
    #   If any method is to be public (either as a @classmethod Post. or instance method post.
    #   it should go in model.py and be in the ModelInterface



    ############################################################################
    ## These query builder passthroughs are for convenience only.  Best to use
    ## MyModel.query().get()... instead (the query() is on the Model.py not
    ## this metaclass).  Why use query()?  Becuase metaclasses are not currently
    ## supported in VSCode for autocomplete code intellisense but parent classes
    ## are.  So by using query() you get autocomplete on the full query builder!
    ############################################################################
    # async def get(entity) -> List[Any]:
    #     """Query builder passthrough"""
    #     return await OrmQueryBuilder(entity).get()

    # async def find(entity, id: Any) -> Any:
    #     """Query builder passthrough"""
    #     return await OrmQueryBuilder(entity).find(id)

    # def where(entity, column: Union[str, List[Tuple]], operator: str = None, value: Any = None):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).where(column, operator, value)

    # def or_where(entity, wheres: List):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).or_where(wheres)

    # def include(entity, *args):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).include(*args)
    ############################################################################
    ############################################################################


    @property
    def pk(entity) -> str:
        """Get the entities primary key"""
        for field in entity.modelfields.values():
            if field.primary: return field.name

    @property
    def connection(entity) -> str:
        """Helper for entity connection string"""
        return entity.__connection__

    @property
    def tablename(entity) -> str:
        """Helper for entity tablename string"""
        return entity.__tablename__

    @property
    def table(entity) -> sa.Table:
        """Helper for entity SQLAlchemy table"""
        return entity.__table__

    @property
    def modelfields(entity) -> Dict[str, Field]:
        """Helper for original uvicore model fields (not pydantic __fields__)"""
        return entity.__modelfields__

    @property
    def modelname(entity) -> str:
        return entity.__name__

    @property
    def modelfqn(entity) -> str:
        module = entity.__module__
        if module is None or module == str.__module__:
            return entity.__name__  # Avoid reporting __builtin__
        else:
            return module + '.' + entity.__name__

    def modelfield(entity, fieldname: str) -> Field:
        field = entity.modelfields.get(fieldname)
        if not field: raise Exception("Field {} not found in model {}".format(fieldname, entity.modelfqn))
        return field

    async def execute(entity, query: Any, values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None) -> sa.CursorResult:
        """Database execute in the context of this entities connection"""
        return await uvicore.db.execute(query=query, values=values, connection=entity.__connection__)

    async def fetchall(entity, query: sa.Select|str, values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None) -> Sequence[sa.Row]:
        """Database fetchall in the context of this entities connection"""
        return await uvicore.db.fetchall(query=query, connection=entity.__connection__)

    async def fetchone(entity, query: sa.Select|str, values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None) -> sa.Row|None:
        """Database fetchone in the context of this entities connection"""
        return await uvicore.db.fetchone(query=query, connection=entity.__connection__)

    # def to_model(entity, row, prefix: str = None) -> Any:
    #     """Convert a row of table data into a model"""
    #     fields = {}
    #     for field in entity.modelfields.values():
    #         if not field.column: continue
    #         column = field.column
    #         if prefix: column = prefix + '.' + column
    #         if hasattr(row, column):
    #             fields[field.name] = getattr(row, column)
    #     return entity(**fields)

    # def to_column(entity, fieldname: str):
    #     """Convert a model field name into a table column name"""
    #     field = entity.modelfields.get(fieldname)
    #     if field: return field.column
    #     return fieldname

    def selectable_columns(entity, table: sa.Table = None, *, show_writeonly: Union[bool, List] = False) -> List[sa.Column]:
        """Get all SQLA columns that are selectable

        Why not just use the table to get all columns?  Because a table
        may have far more columns than the actual model.  So we use the model
        to infer a list of actual SQLA columns (excluding write_only fields)
        """
        if table is None: table = entity.table

        # Not all models require tables (databaseless models)
        if table is None: return []

        all_columns = table.columns
        columns: List[sa.Column] = []
        for (field_name, field) in entity.modelfields.items():
            # Exclude None columns (which are relations) and write_only columns which cannot be viewed
            if field.column:
                show = False
                if not field.write_only: show = True
                if type(show_writeonly) == bool:
                    # show_writeonly is a bool, meaning show all writeonly fields
                    if field.write_only and show_writeonly == True: show = True
                else:
                    # show_writeonly is a list of fields to allow
                    if field.write_only and field.column in show_writeonly: show = True

                if show:
                    columns.append(getattr(all_columns, field.column))
        return columns

    def info(entity) -> Dict[str, Any]:
        fields = {}
        for (field_name, field) in entity.modelfields.items():
            fields[field_name] = field
        return {
            'connection': uvicore.db.connection(entity.__connection__),
            'tablename': entity.tablename,
            'table': entity.table,
            'fields': fields,
        }

    def __new__(mcls: type, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any], **kwargs) -> type:
        # mcls is this ModelMetaclass itself
        # name is the string name of the child class, in this case '_Model' from model.py
        # bases is a tuple of parent used in the child (_Model) class, in this case (uvicore.contracts.model.Model, pydantic.main.BaseModel) it does not include this metaclass
        # namespace is the child _Model classes original __dict__ Dictionary

        # Define our custom attributes
        __connection__ = None
        __tablename__ = None
        __table__ = None
        __tableclass__ = None
        __callbacks__ = {}

        # Pull out all model properties of type Field() and store in __modelfields__ property
        # Then replace the original properties from Field() to Pydantics FieldInfo(), converting some Field() arguments
        # into 'x-tra' arguments for FieldInfo()
        # In the end, __modelfields__ are the original uvicore Field() that you defined in the model
        # and the original model fields are replaced with pydantics FieldInfo().  Later when I call super().__new__
        # pydantic further converts by MOVING the actual class properties into __fields__ and changing from FieldInfo to ModelField
        __modelfields__: Dict[str, Field] = {}
        for field_name, field in namespace.items():
            if field_name[0] != '_' and type(field) == Field:
                # Pull out uvicore model Field() into own __modelfields__ dict
                field.name = field_name
                __modelfields__[field_name] = field

                # Derive extra field information for pydantics FieldInfo OpenAPI schema generation
                #dump(getattr(namespace, field_name))

                # Convert uvicore model field into pydantic FieldInfo
                # Be careful what you pipe into PydanticFieldInfo() as these show up in the OpenAPI schema
                # from FastAPI 0.65.0+ because in their openapi/models.py they set class Config extra='allow'
                # by default.  Which means any item you put in PydanticFieldInfo that is not a keyword on that class
                # goes into extra *kwargs and extra now shows up in openapi.json.  The problem with that is valid
                # OpenAPI only allows certain keywords, see https://swagger.io/docs/specification/data-models/keywords/
                # In fact, the 'properties' I was using for a catch-all JSON blog is totally invalid.  Its not a ignored blog
                # its a nested schame field properties.  The only valid way to add extra ignored fields to a schema is to
                # prefix them with x-  see https://swagger.io/specification/#specification-extensions

                # For some dumb reason I was also piping EVERY thing into PydanticFieldInfo, even callback, evaluate
                # and relation.  Which now that extra=allow those were all trying to be parsed by FastAPI json_encoder
                # which totally broke everything.  So I need to be more strategic which uvicore Field() items I pump into
                # Pydantic FieldInfo().  I added a fields_passed_to_pydantic_FieldInfo in my Field class that contains a
                # List of valid properties to pass into PydanticFieldInfo()

                field_info_kwargs = {}
                #dump(field.__annotations__)
                for slot in field.__annotations__.keys():
                    value = getattr(field, slot)
                    if value is None: continue;

                    # Only pipe certain uvicore Field properties into pydantics FieldInfo or you get invalid OpenAPI spec
                    if slot in Field.__valid_oepnapi_keywords__:
                        arg = slot

                        # Convert some understores to camelCase for OpenAPI keyword compatibility
                        if slot == 'read_only': arg = 'readOnly'
                        if slot == 'write_only': arg = 'writeOnly'
                        if slot == 'min_length': arg = 'minLength'
                        if slot == 'max_length': arg = 'maxLength'
                        field_info_kwargs[arg] = value

                    elif slot in Field.__convert_to_extensions__:
                        # Convert these Field() arguments to the x-tra Dict
                        # See https://swagger.io/specification/#specification-extensions
                        if 'x-tra' not in field_info_kwargs: field_info_kwargs['x-tra'] = {}
                        if slot == 'properties':
                            field_info_kwargs['x-tra'] = {**field_info_kwargs['x-tra'], **getattr(field, slot)}
                        else:
                            field_info_kwargs['x-tra'][slot] = getattr(field, slot)

                #dump(field_info_kwargs)
                namespace[field_name] = PydanticFieldInfo(**field_info_kwargs)
                #dump(namespace[field_name])

        #dump(namespace)

        # If we extend and overwrite our own models, then some information
        # will be buried in the bases tuple.  Loop each base and
        # pluck out these critical fields (for modelfields, APPEND them to allow extension)
        for base in bases:
            if hasattr(base, '__modelfields__'):
                # Notice we are APPEND modelfields to allow model extension
                __modelfields__ = {**base.__modelfields__, **__modelfields__}

            # I am only setting these if not already set.  This allows a higher base
            # like the parent to override and win over the base children
            if hasattr(base, '__connection__') and not __connection__:
                __connection__ = base.__connection__
            if hasattr(base, '__tablename__') and not __tablename__:
                __tablename__ = base.__tablename__
            if hasattr(base, '__table__') and not __table__:
                __table__ = base.__table__
            if hasattr(base, '__tableclass__') and not __tableclass__:
                __tableclass__ = base.__tableclass__
            if hasattr(base, '__callbacks__') and not __callbacks__:
                __callbacks__ = base.__callbacks__

        # Add my own ORM attributes to pydantics base ModelMetaClass
        new_namespace = {
            '__connection__': __connection__,
            '__tablename__': __tablename__,
            '__table__': __table__,
            '__tableclass__': __tableclass__,
            '__callbacks__': __callbacks__,
            '__modelfields__': __modelfields__,
            #'__query__': {},
            #'_test1': 'hi',
            **{n: v for n, v in namespace.items()},
        }

        #dd(new_namespace)

        # Call pydantic ModelMetaClass
        # Amoung other things, pydantic will take all model attributes that do not
        # begin with a _ and convert them into ModelField classes
        # This is why I keep the originals in my new __modelfields__ attribute
        cls = super().__new__(mcls, name, bases, new_namespace, **kwargs)
        #dump(name, cls, cls.__dict__, '-----------------------------')

        # Meta is fired up more than once, sometimes pydantic has NOT
        # actually populated all fields.  If no fields, ignore rest of this custom __new__
        if not cls.__fields__: return cls

        # Pretty Printer
        # Register a pretty printer just for this entity.  Why not on the main
        # Model class itself?  Because then it prints each record as if it were
        # uvicore.orm.model.Model instead of the actual model class (ie: uvicore.auth.models.user.User)
        @register_pretty(cls)
        def pretty_entity(value, ctx):
            return pretty_call(ctx, cls, **value.__dict__)

        #dump("Registering Schema in Metaclass")

        # Build connection, tablename and table from tableclass
        if cls.__tableclass__ is not None:
            if cls.__connection__ is None: cls.__connection__ = cls.__tableclass__.connection
            if cls.__tablename__ is None: cls.__tablename__ = cls.__tableclass__.name
            if cls.__table__ is None: cls.__table__ = cls.__tableclass__.schema


        # Dynamically Build SQLAlchemy Table From Model Properties
        if cls.__table__ is not None:
            pass
            # Set connection and table name from table class
            #dd(cls.__table__.metadata.__dict__)
            #dd(dir(cls.__table__))
            #if cls.__connection__ is None: cls.__connection__ = cls.__table__.connection
            #if cls.__tablename__ is None: cls.__tablename__ = cls.__table__.name

        else:
            pass
            #dump('Building SA Table From Model Properties')
            # Fixme, code could go here to dynamically build
            # an SQLAlchemy table here.

            # An example of basic build, but fix to use __fields__
            # and each fields .field_info instead, and convert each model
            # property Field() into SA table stuff like columns and indexes...
            # columns = [x for x in entity.__field_defaults__.values()]
            # __class__.__table__ = sa.Table(
            #     entity.__tablename__,
            #     db.metadata.get(entity.__connection__),
            #     *columns
            # )

        # Pull out all callbacks from __modelfields__ and store in cls.__callbacks__ for future processing
        for (key, field) in cls.__modelfields__.items():
            if field.callback:
                callback = field.callback
                if type(callback) == str:
                    callback = getattr(cls, field.callback)
                cls.__callbacks__[key] = callback

        #dump(cls.__dict__)
        return cls


# IoC Class Instance
#ModelMetaclass: _ModelMetaclass = uvicore.ioc.make('ModelMetaclass', _ModelMetaclass)



import uvicore
from uvicore.support import module
from uvicore.typing import Dict, List
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Email


@uvicore.service()
class Mail:

    def __init__(self, *,
        mailer: str = None,
        mailer_options: Dict = None,
        to: List = [],
        cc: List = [],
        bcc: List = [],
        from_name: str = None,
        from_address: str = None,
        subject: str = None,
        html: str = None,
        text: str = None,
        attachments: List = [],
    ) -> None:
        # Get mailer and options from config
        self._config = uvicore.config.app.mail.clone()
        self._mailer = mailer or self._config.default
        self._mailer_options = self._config.mailers[self._mailer].clone().merge(mailer_options)

        # New message superdict
        self._message: Email = Email()
        self._message.to = to
        self._message.cc = cc
        self._message.bcc = bcc
        self._message.from_name = from_name or self._config.from_name
        self._message.from_address = from_address or self._config.from_address
        self._message.subject = subject
        self._message.html = html
        self._message.text = text
        self._message.attachments = attachments

    def mailer(self, mailer: str):
        self._mailer = mailer
        self._mailer_options = self._config.mailers[self._mailer].clone()
        return self

    def mailer_options(self, options: Dict):
        self._mailer_options.merge(Dict(options))
        return self

    def to(self, to: List):
        self._message.to = to
        return self

    def cc(self, cc: List):
        self._message.cc = cc
        return self

    def bcc(self, bcc: List):
        self._message.bcc = bcc
        return self

    def from_name(self, from_name: str):
        self._message.from_name = from_name
        return self

    def from_address(self, from_address: str):
        self._message.from_address = from_address
        return self

    def subject(self, subject: str):
        self._message.subject = subject
        return self

    def html(self, html: str):
        self._message.html = html
        return self

    def text(self, text: str):
        self._message.text = text
        return self

    def attachments(self, attachments: List):
        self._message.attachments = attachments
        return self

    async def send(self):
        # Use dynamic module based on mailer driver
        driver = module.load(self._mailer_options.driver).object
        await driver.send(self._message, self._mailer_options)



from __future__ import annotations

import re
import sys
import uvicore
import logging
import logging.config
from logging import Formatter
from uvicore.typing import List
from colored import attr, bg, fg
from logging import Logger as PythonLogger
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Logger as LoggerInterface


class OutputFilter(logging.Filter):
    """Python logging custom filter class"""

    def __init__(self, filters, excludes):
        self.filters = filters
        self.excludes = excludes
        super().__init__(name='')

    def filter(self, record):
        # Not an exact filter match but a contains match.  This matches how default python
        # logging filters are.  So you can filter on A.B and it will include
        # names of A.B.C and up.
        show = False
        if self.filters:
            for f in self.filters:
                if record.name[0:len(f)] == f:
                    show = True
                    break
        else:
            show = True

        if show and self.excludes:
            for exclude in self.excludes:
                if record.name[0:len(exclude)] == exclude:
                    show = False
                    break

        return show



class ExcludeFilter(logging.Filter):
    """Python logging custom exclude filter class"""

    def __init__(self, excludes):
        self.excludes = excludes
        super().__init__(name='exclude')

    def filter(self, record):
        # Not an exact filter match but a contains match.  This matches how default python
        # logging filters are.  So you can filter on A.B and it will include
        # names of A.B.C and up.
        for exclude in self.excludes:
            print(exclude)
            if record.name[0:len(exclude)] == exclude: return False
        return True


class ColoredFormatter(Formatter):

    def __init__(self, patern):
        Formatter.__init__(self, patern)

    def format(self, record):
        # Remember this is console output only, not file or other handlers
        # See color chart https://pypi.org/project/colored/
        level = record.levelname
        message = logging.Formatter.format(self, record)
        prefix = str(message.strip()[0:2]).strip()

        # Format all INFO level messages
        if level == 'INFO':

            # Format header
            if prefix == '::':
                message = re.sub("^:: ", "", message)
                message = re.sub(" ::$", "", message)
                message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ':: ', attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ::', attr(0))

            # Format header2
            if prefix == '##':
                message = re.sub("^## ", "", message)
                message = re.sub(" ##$", "", message)
                message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '## ', attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ##', attr(0))

            # Format header3
            if prefix == '==':
                message = re.sub("^=== ", "", message)
                message = re.sub(" ===$", "", message)
                message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '=== ', attr(0)) \
                    + ('{0}{1}{2}').format(fg('green'), message, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ===', attr(0))

            # Format header4
            if prefix == '--':
                message = re.sub("^---- ", "", message)
                message = re.sub(" ----$", "", message)
                message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '---- ', attr(0)) \
                    + ('{0}{1}{2}').format(fg('dark_green'), message, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ----', attr(0))

            # Format item *
            elif prefix == '*':
                split = message.split('*')
                pre = split[0] + '*'
                post = '*'.join(split[1:])
                message = ('{0}{1}{2}').format(fg('green'), pre, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), post, attr(0))

            # Format item2 -
            elif prefix == '-':
                split = message.split('-')
                pre = split[0] + '-'
                post = '-'.join(split[1:])
                message = ('{0}{1}{2}').format(fg('red'), pre, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), post, attr(0))

            # Format item3 +
            elif prefix == '+':
                split = message.split('+')
                pre = split[0] + '+'
                post = '+'.join(split[1:])
                message = ('{0}{1}{2}').format(fg('cyan'), pre, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), post, attr(0))

            # Format item4 >
            elif prefix == '>':
                split = message.split('>')
                pre = split[0] + '>'
                post = '>'.join(split[1:])
                message = ('{0}{1}{2}').format(fg('magenta'), pre, attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), post, attr(0))

            # Format notice
            elif (level == 'INFO' and re.match("^NOTICE: ", message)):
                message = re.sub("^NOTICE: ", "", message)
                message = ('{0}{1}{2}{3}').format(fg('yellow'), attr('bold'), 'NOTICE: ', attr(0)) \
                    + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

            # Format separator
            elif (level == 'INFO' and re.match("^====", message)):
                message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

            # Format line
            elif (level == 'INFO' and re.match("^----", message)):
                message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

            # No special formatting, plain old .info()
            else:
                message = message

        elif (level == 'DEBUG'):
            message = ('{0}{1}{2}').format(fg(241), message, attr(0))

        elif (level == 'WARNING'):
            message = ('{0}{1}{2}').format(fg('orange_red_1'), message, attr(0))
        elif (level == 'ERROR'):
            message = ('{0}{1}{2}').format(fg('red'), message, attr(0))
        elif (level == 'CRITICAL'):
            message = ('{0}{1}{2}{3}').format(fg('black'), bg('red'), message, attr(0))

        return message


@uvicore.service('uvicore.logging.logger.Logger',
    aliases=['Logger', 'logger', 'Log', 'log'],
    singleton=True,
    kwargs={'config': uvicore.config('app.logger')},
)
class Logger(LoggerInterface):
    """Logger private class.

    Do not import from this location.
    Use the uvicore.log singleton global instead."""

    def __init__(self, config):
        # Default Config
        # Levels from logging._levelToName are
        # {50: 'CRITICAL', 40: 'ERROR', 30: 'WARNING', 20: 'INFO', 10: 'DEBUG', 0: 'NOTSET'}

        # Levels = DEBUG, INFO, WARNING, ERROR, CRITICAL
        default = {
            'console': {
                'enabled': True,
                'level': 'DEBUG',
                'colors': True,
                'format': '%(message)s',
                'filters': [],
                'exclude': [],
            },
            'file': {
                'enabled': False,
                'level': 'DEBUG',
                'file': '/tmp/example.log',
                'when': 'midnight',
                'interval': 1,
                'backup_count': 7,
                'format': '%(asctime)s.%(msecs)03d | %(levelname)-8s | %(name)-22s | %(message)s',
                'filters': [],
                'exclude': [],
            }
        }

        # Merge default and user defined config
        config = {**default, **config}
        if 'console' in config.keys(): config['console'] = {**default['console'], **config['console']}
        if 'file' in config.keys(): config['file'] = {**default['file'], **config['file']}

        # New Logger
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        self._name = None

        # New Console Handler
        if config['console']['enabled']:
            handler = logging.StreamHandler(stream=sys.stdout)
            handler.setLevel(config['console']['level'])
            if config['console']['colors']:
                handler.setFormatter(ColoredFormatter(config['console']['format']))
            else:
                handler.setFormatter(logging.Formatter(
                    fmt=config['console']['format'],
                    datefmt='%Y-%m-%d %H:%M:%S'
                ))
            handler.addFilter(OutputFilter(config['console']['filters'], config['console']['exclude']))
            self._logger.addHandler(handler)

        # New File Handler
        if config['file']['enabled']:
            #class logging.handlers.TimedRotatingFileHandler(filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False, atTime=None, errors=None)
            #handler = logging.FileHandler(filename=config['file']['file'], mode='a')
            handler = logging.handlers.TimedRotatingFileHandler(filename=config['file']['file'], when=config['file']['when'], interval=config['file']['interval'], backupCount=config['file']['backup_count'])
            handler.setLevel(config['file']['level'])
            handler.setFormatter(logging.Formatter(
                fmt=config['file']['format'],
                datefmt='%Y-%m-%d %H:%M:%S'
            ))
            #if config['file'].get('filter'): handler.addFilter(logging.Filter(name=config['file']['filter']))
            handler.addFilter(OutputFilter(config['file']['filters'], config['file']['exclude']))
            self._logger.addHandler(handler)

        self.config = config

    def __call__(self, message):
        self.info(message)

    @property
    def console_handler(self) -> PythonLogger:
        try:
            return self._logger.handlers[0]
        except IndexError:
            return None

    @property
    def file_handler(self) -> PythonLogger:
        try:
            return self._logger.handlers[1]
        except IndexError:
            return None

    @property
    def logger(self):
        if not self._name: return self._logger
        return logging.getLogger(self._name)

    def name(self, name: str) -> LoggerInterface:
        self._name = name
        return self

    def reset(self):
        self._name = None

    def dump(self, *args):
        running_pytest = uvicore.app.is_pytest
        console_enabled = self.config['console']['enabled']
        console_level = logging.getLevelName(uvicore.log.console_handler.level)
        console_filters = self.config['console']['filters']
        console_excludes = self.config['console']['exclude']
        file_enabled = self.config['file']['enabled']
        file_level = logging.getLevelName(uvicore.log.file_handler.level)

        # Use dump() to prettyprint to console only if console is in DEBUG mode or we are running a pytest.
        # The dump() does not understand log filters and excludes, so we must use those manually to decide
        # if we should dump() the content or not.
        if (console_enabled and console_level == 'DEBUG') or running_pytest:
            show = False
            loggerName = self._name or 'root'

            # Check filters
            if not console_filters: show = True
            if not show:
                for filter in console_filters:
                    if loggerName[0:len(filter)] == filter:
                        show = True
                        break

            # Check excludes
            if show and console_excludes:
                for exclude in console_excludes:
                    if loggerName[0:len(exclude)] == exclude:
                        show = False
                        break

            # Loglevel, Filters and Excludes say we can dump this to the console
            if show: dump(*args)

        # Dump to file
        if (file_enabled and file_level == 'DEBUG'):
            # We must temporarily disable the console logger or this prints to the console as well
            # Which means a double print because of the dump(*args) abvove
            # Must get handlers from the root logger.  If we are using a custom logger name like 'my-section' then it has no handlers
            # so self.logger.handlers does not work.  We want the 'root' logger handlers
            root_handlers = logging.getLogger().handlers
            for handler in root_handlers:
                if 'logging.StreamHandler' in str(type(handler)):
                    handler.setLevel('CRITICAL')

            # Log to file in in DEBUG mode
            for arg in args:
                self.logger.debug(arg)

            # Re-enable console logger by restoring the original level
            for handler in root_handlers:
                if 'logging.StreamHandler' in str(type(handler)):
                    handler.setLevel(console_level)

        # Reset logger name
        self.reset()

    def info(self, message):
        self.logger.info(str(message))
        self.reset()

    def notice(self, message):
        self.logger.info("NOTICE: " + str(message))
        self.reset()

    def warning(self, message):
        self.logger.warning(str(message))
        self.reset()

    def debug(self, message):
        self.logger.debug(str(message))
        self.reset()

    def error(self, message):
        self.logger.error(str(message))
        self.reset()

    def critical(self, message):
        self.logger.critical(str(message))
        self.reset()

    def exception(self, message):
        self.logger.error(str(message))
        self.reset()

    def blank(self):
        self.logger.info('')
        self.reset()

    def nl(self) -> LoggerInterface:
        """nl() is a blank() that is chainable"""
        self.logger.info('')
        return self

    def separator(self):
        self.logger.info('=' * 80)
        self.reset()

    def line(self):
        self.logger.info('-' * 80)
        self.reset()

    def header(self, message):
        self.logger.info(":: " + str(message) + " ::")
        self.reset()

    def header2(self, message):
        self.logger.info("## " + str(message) + " ##")
        self.reset()

    def header3(self, message):
        self.logger.info("=== " + str(message) + " ===")
        self.reset()

    def header4(self, message):
        self.logger.info("---- " + str(message) + " ----")
        self.reset()

    def item(self, message, *, level: int = 1):
        spaces = ' ' * (level * 4)
        self.logger.info(spaces + "* " + str(message))
        self.reset()

    def item2(self, message, *, level: int = 1):
        spaces = ' ' * (level * 4)
        self.logger.info(spaces + "- " + str(message))
        self.reset()

    def item3(self, message, *, level: int = 1):
        spaces = ' ' * (level * 4)
        self.logger.info(spaces + "+ " + str(message))
        self.reset()

    def item4(self, message, *, level: int = 1):
        spaces = ' ' * (level * 4)
        self.logger.info(spaces + "> " + str(message))
        self.reset()



# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.log singleton global instead.

# Public API for import * and doc gens
#__all__ = ['_Logger', 'ColoredFormatter']



import uvicore
from fastapi import FastAPI as _FastAPI
from uvicore.support.dumper import dd, dump


@uvicore.service('uvicore.http.servers.api._Server',
    #aliases=['api', 'web_server'],
    singleton=True,
    kwargs={
        'debug': uvicore.config('app.debug'),
        'title': uvicore.config('app.api.openapi.title'),
        'version': uvicore.app.version,
        'openapi_url': uvicore.config('app.api.openapi.url'),
        'docs_url': uvicore.config('app.api.openapi.docs_url'),
        'redoc_url': uvicore.config('app.api.openapi.redoc_url'),
        #'root_path': uvicore.config('app.api.root_path'), # experiment with using Kong for API but Nginx for Web, may need 2 root_paths
    },
)
class Server(_FastAPI):
    @property
    def server(self) -> _FastAPI:
        return self



import uvicore
from uvicore.support.dumper import dump, dd


@uvicore.seeder()
async def seed():
    # Import seeders
    from . import users, groups, roles, permissions

    # Run seeders. Order is critical for ForeignKey dependencies
    await permissions.seed()
    await roles.seed()
    await groups.seed()
    await users.seed()



import json
from fastapi.security import SecurityScopes
from fastapi.params import Security as SecurityClass
from fastapi.datastructures import UploadFile, DefaultPlaceholder
from starlette.requests import Request, HTTPConnection, ClientDisconnect, cookie_parser

# NOTE, all params are also available at uvicore.http.params
# Difference between fastapi.params vs fastapi.param_functions??
# param_functions add more props and defaults then instantiate
# the param Class themselves.  Like a higher wrapper around the classes.
# All docs say 'from fastapi import Form' which actually comes from param_functions
from fastapi.param_functions import Path
from fastapi.param_functions import Query
from fastapi.param_functions import Header
from fastapi.param_functions import Cookie
from fastapi.param_functions import Body
from fastapi.param_functions import Form
from fastapi.param_functions import File
from fastapi.param_functions import Depends
from fastapi.param_functions import Security

class Parameter(SecurityClass):
    """Base class for user defined request parameters with infinite kwargs!"""

    # Parameters can accept any number of kwargs which are hacked as
    # a single "string" security scope converted to json then converted
    # back to an object before the handler is called

    #def __init__(self, scopes: Sequence[str] = []):
    def __init__(self, **kwargs):
        # Hack, scopes must be strings, convert kwargs into one JSON string
        scopes = [json.dumps(kwargs)]

        # Call FastAPI Security Depends with "self" and one scope as a JSON string
        super().__init__(dependency=self, scopes=scopes, use_cache=True)

    async def __call__(self, security_scopes: SecurityScopes, request: Request):
        # Hack, scopes are a JSON blog of kwargs
        kwargs = json.loads(security_scopes.scopes[0])

        # Call user defined handler
        return await self.handle(request=request, **kwargs)



import uvicore
from uvicore.typing import Dict
from uvicore.events import Handler
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Package as Package
from uvicore.foundation.events.app import Booted as OnAppBooted


class Templating(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the Templating System after the Application is Booted"""

        # Initialize Templating System
        self.initialize_templates()


    def initialize_templates(self) -> None:
        """Initialize Templating System"""

        # Loop each package with an HTTP definition and add to our HTTP server
        paths = []
        context_processors = Dict()
        for package in uvicore.app.packages.values():
            if not 'templating' in package: continue

            # Append template paths
            for path in package.templating.paths or []:
                paths.append(path)

            # Deep merge template context_processors
            context_processors.merge(package.templating.context_processors or {})

        # Get the template singleton from the IoC
        templates = uvicore.ioc.make('uvicore.templating.engine.Templates')

        # Add all packages view paths
        # First path wins, so we must REVERSE the package order
        # This will set the main running app FIRST as we always want the app to win
        paths.reverse()
        for path in paths:
            if '/' in path:
                # Path is an actual unix path
                templates.include_path(path)
            else:
                # Path is probably a python myapp.template path, convert to unix path
                templates.include_path(module.location(path))
        # Debug, can see actual jinja paths like so
        # dd(templates._paths)

        # Add all packages deep_merged template options
        if 'context_functions' in context_processors:
            for name, method in context_processors['context_functions'].items():
                templates.include_context_function(name, method)
        if 'context_filters' in context_processors:
            for name, method in context_processors['context_filters'].items():
                templates.include_context_filter(name, method)
        if 'filters' in context_processors:
            for name, method in context_processors['filters'].items():
                templates.include_filter(name, method)
        if 'tests' in context_processors:
            for name, method in context_processors['tests'].items():
                templates.include_test(name, method)

        # Initialize template system
        templates._init()



from uvicore.typing import Any, Callable
from abc import ABC, abstractmethod, abstractproperty
from uvicore.typing import Dict, List, Any, Union, Callable


class JobDispatcher(ABC):
    @abstractmethod
    def dispatch(self, instance: object) -> Any:
        """Dispatch a Job Class"""
        pass

    @abstractmethod
    async def dispatch_async(self, instance: object) -> Any:
        """Dispatch an async Job Class"""
        pass

    @abstractmethod
    async def codispatch(self, instance: object) -> Any:
        """Alias for dispatch_async()"""
        pass




#import typer_async as typer
from typing import Optional

import uvicore
from uvicore import app, db, log
from uvicore.console import argument, click, command, option
from uvicore.support import module
from uvicore.support.dumper import dd, dump

# Commands
# create = typer.Typer()
# drop = typer.Typer()
# recreate = typer.Typer()
# seed = typer.Typer()
# reseed = typer.Typer()
# connections = typer.Typer()

# Common arguments
#connection_arg = typer.Argument(..., help='Connection(s) name (comma separated for multiples)')

def get_metakeys(connections: str):
    """Convert connections into deduplicated metakeys"""
    metakeys = []
    connections = connections.split(',') if ',' in connections else [connections]
    for connection in connections:
        conn = db.connection(connection)
        if conn:
            metakey = conn.metakey
            if metakey not in metakeys:
                metakeys.append(metakey)
        else: # pragma: no cover
            exit('Connection string {} not found in configs'.format(connection))
    return metakeys


async def create_tables(connections: str):
    """Create tables for connection(s)"""

    # Convert connections into List
    connections = connections.split(',') if ',' in connections else [connections]
    for conn_str in connections:
        # Get actual connection configuration Dict
        connection = db.connection(conn_str)

        # Get [sync or async] engine
        engine = db.engine(metakey=connection.metakey)

        # Get metadata (table information) for this connection
        metadata = db.metadata(metakey=connection.metakey)

        # Log
        log.header('Creating tables in {} in topological order'.format(connection.metakey))
        for table in metadata.sorted_tables:
            log.item('Creating table {}'.format(str(table.name)))

        if connection.is_async:
            # Create tables with async driver
            async with engine.begin() as conn:
                await conn.run_sync(metadata.create_all)
        else:
            # Create tables with sync driver
            metadata.create_all(engine)
        print()


async def drop_tables(connections: str):
    """Drop tables for connection(s)"""

    # Convert connections into List
    connections = connections.split(',') if ',' in connections else [connections]
    for conn_str in connections:
        # Get actual connection configuration Dict
        connection = db.connection(conn_str)

        # Get [sync or async] engine
        engine = db.engine(metakey=connection.metakey)

        # Get metadata (table information) for this connection
        metadata = db.metadata(metakey=connection.metakey)

        # Log
        log.header('Dropping tables from {} in topological order'.format(connection.metakey))
        for table in reversed(metadata.sorted_tables):
            log.item('Dropping table {}'.format(str(table.name)))

        if connection.is_async:
            # Drop tables with async driver
            async with engine.begin() as conn:
                await conn.run_sync(metadata.drop_all)
        else:
            # Drop tables with sync driver
            metadata.drop_all(engine)
        print()


async def seed_auth_permissions():
    if 'auth' not in db.connections().keys(): return

    from uvicore.auth.models.permission import Permission
    metadata = db.metadata('auth')

    # Default permissions per entity
    permissions = ['create', 'read', 'update', 'delete']

    # Bulk insert each permission for each table
    bulk = []
    tables = sorted(metadata.tables.keys())  # For sorted order
    for table in tables:
        for permission in permissions:
            bulk.append(Permission(entity=table, name=table + '.' + permission))
    await Permission.insert(bulk)


async def seed_tables(connections: str):
    metakeys = get_metakeys(connections)

    # Seed auth permissions
    await seed_auth_permissions()

    # Seed tables
    ran = []
    for metakey in metakeys:
        packages = db.packages(metakey=metakey)
        for package in packages:
            for seeder in package.database.seeders:
                if seeder not in ran:
                    log.header('Seeding tables from {} in defined order'.format(seeder))
                    await module.load(seeder).object()
                    ran.append(seeder)
                    print()


async def reseed_tables(connection:str):
    await drop_tables(connection)
    await create_tables(connection)
    await seed_tables(connection)


@command()
@argument('connections')
async def create(connections: str): # pragma: no cover
    """Create tables for connection(s)"""
    await create_tables(connections)


@command()
@argument('connections')
async def drop(connections: str): # pragma: no cover
    """Drop tables for connection(s)"""
    await drop_tables(connections)


@command()
@argument('connections')
async def recreate(connections: str): # pragma: no cover
    """Recreate (drop/create) tables for connection(s)"""
    #log.header('Recreating (drop/create) tables for connections: [' + connections + ']')
    await drop_tables(connections)
    await create_tables(connections)


@command()
@argument('connections')
async def seed(connections: str): # pragma: no cover
    """Seed tables for connection(s)"""
    await seed_tables(connections)


@command()
@argument('connections')
async def reseed(connections: str): # pragma: no cover
    """Reseed (drop/create/seed) tables for connection(s)"""
    await reseed_tables(connections)


@command()
def connections(): # pragma: no cover
    """Show all packages database connections"""
    log.header("Final Merged Database Connections")
    log.line()
    log.notice("Some connections share the same database which means their tables are in the same metedata space.  This is what the unique metakey denotes.")
    dump(db.connections)
























# @create.command()
# def create_cmd(connections: str = connection_arg):
#     """Create tables for connection(s)"""
#     #log.header('Creating tables for connections: [' + connections + ']')
#     metakeys = get_metakeys(connections)
#     for metakey in metakeys:
#         log.header('Creating tables in {} in topologically order'.format(metakey))
#         metadata = db.metadata(metakey=metakey)
#         for table in metadata.sorted_tables:
#             log.item('Creating table {}'.format(str(table.name)))
#         metadata.create_all(db.engine(metakey=metakey))
#         print()

# @drop.command()
# def drop_cmd(connections: str = connection_arg):
#     """Drop tables for connection(s)"""
#     #log.header('Dropping tables for connections: [' + connections + ']')
#     metakeys = get_metakeys(connections)
#     for metakey in metakeys:
#         log.header('Dropping tables from {} in topologically order'.format(metakey))
#         metadata = db.metadata(metakey=metakey)
#         for table in reversed(metadata.sorted_tables):
#             log.item('Dropping table {}'.format(str(table.name)))
#         metadata.drop_all(db.engine(metakey=metakey))
#         print()

# @recreate.command()
# def recreate_cmd(connections: str = connection_arg):
#     """Recreate (drop/create) tables for connection(s)"""
#     #log.header('Recreating (drop/create) tables for connections: [' + connections + ']')
#     drop_cmd(connections)
#     create_cmd(connections)

# @seed.command()
# def seed_cmd(connections: str = connection_arg):
#     """Seed tables for connection(s)"""
#     metakeys = get_metakeys(connections)
#     ran = []
#     for metakey in metakeys:
#         packages = db.packages(metakey=metakey)
#         for package in packages:
#             for seeder in package.seeders:
#                 if seeder not in ran:
#                     log.header('Seeding tables from {} in defined order'.format(seeder))
#                     module.load(seeder).object()
#                     ran.append(seeder)
#                     print()

# @reseed.command()
# def reseed_cmd(connections: str = connection_arg):
#     """Reseed (drop/create/seed) tables for connection(s)"""
#     recreate_cmd(connections)
#     seed_cmd(connections)

# @connections.command()
# def connections_cmd():
#     """Show all packages database connections"""
#     log.header("All deep merged database connections from all defined packages").line()
#     log.notice("Some connections share the same database which means their tables are in the same metedata space.  This is what the unique metakey denotes.")
#     dump(db.connections)



from abc import ABC, abstractmethod
from uvicore.typing import Dict, List, Any, Union, Callable


class Dispatcher(ABC):

    # @property
    # @abstractmethod
    # def events(self) -> Dict[str, Dict]:
    #     """Dictionary of all registered events in uvicore and all packages"""
    #     pass

    @property
    @abstractmethod
    def listeners(self) -> Dict[str, List]:
        """Dictionary of all listeners for each event"""
        pass

    @property
    @abstractmethod
    def wildcards(self) -> List:
        """List of all wildcard listeners"""
        pass

    # @abstractmethod
    # def event(self, event: Union[str, Callable]) -> Dict:
    #     """Get one event by str name or class"""
    #     pass

    @abstractmethod
    def event_listeners(self, event: str) -> List:
        """Get all listeners for an event including wildcard, sorted by priority ASC"""
        pass

    # @abstractmethod
    # def register(self, events: Dict[str, Dict] = None, *, name: str = None, description: str = None, is_async: bool = False, dynamic: bool = True):
    #     """Register an event with the system.  Retrieve with .events property"""
    #     pass

    @abstractmethod
    def listen(self, events: Union[str, List], listener: Union[str, Callable] = None, *, priority: int = 50) -> None:
        """Append a listener (string or method) callback to one or more events"""
        pass

    def handle(self, events: Union[str, List], listener: Union[str, Callable] = None, *, priority: int = 50) -> None:
        """Alias to listen"""
        pass

    @abstractmethod
    def subscribe(self, listener: Union[str, Callable]) -> None:
        """Add a subscription class which handles both registration and listener callbacks"""
        pass

    @abstractmethod
    def dispatch(self, event: Any, payload = {}) -> None:
        """Fire off an event and run all listener callbacks"""
        pass

    @abstractmethod
    async def dispatch_async(self, event: Any, payload = {}) -> None:
        """Fire off an event and run all async listener callbacks"""
        pass

    @abstractmethod
    async def codispatch(self, event: Any, payload: Dict = {}) -> None:
        """Alias for dispatch_async()"""
        pass




import uuid
import uvicore
from uvicore.auth.models.user import User
from uvicore.auth.models.role import Role
from uvicore.support.dumper import dump, dd


@uvicore.seeder()
async def seed():
    uvicore.log.item('Seeding table users')

    # Get all roles keyed by name
    roles = await Role.query().key_by('name').get()

    unique_id = str(uuid.uuid4())
    # How you "would" do it, but None password means login never allowed
    #pwd = password.create(unique_id)
    pwd = None

    user = await User(
        uuid=unique_id,
        username='anonymous',
        email='anonymous@example.com',
        first_name='Anonymous',
        last_name='User',
        title='Anonymous',
        disabled=True,
        password=None,
        creator_id=1,
    ).save()

    await user.link('roles', [
        roles['Anonymous']
    ])

    # users = []
    # fake = Faker()
    # for _ in range(2):
    #     user = User(
    #         email=fake.email()
    #     )
    #     await user.save()



import uvicore
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.http.package.registers import Http
from uvicore.database.package.registers import Db


@uvicore.provider()
class Auth(Provider, Db, Http):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register IoC bindings
        # self.bind(
        #     name='Auth',
        #     object='uvicore.auth.auth._Auth',
        #     aliases=['auth']
        # )

        # Bind Tables
        #self.bind('uvicore.auth.database.tables.groups.Groups', 'uvicore.auth.database.tables.groups._Groups', singleton=True)
        #self.bind('uvicore.auth.database.tables.user_info.UserInfo', 'uvicore.auth.database.tables.user_info._UserInfo', singleton=True)
        #self.bind('uvicore.auth.database.tables.users.Users', 'uvicore.auth.database.tables.users._Users', singleton=True)

        # Bind Models
        #self.bind('uvicore.auth.models.group.Group', 'uvicore.auth.models.group.GroupModel')
        #self.bind('uvicore.auth.models.user.User', 'uvicore.auth.models.user.UserModel')
        #self.bind('uvicore.auth.models.user_info.UserInfo', 'uvicore.auth.models.user_info.UserInfoModel')

        # Register config
        self.configs([
            {'key': self.name, 'module': 'uvicore.auth.config.package.config'}
        ])

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define Database Connections
        self.register_db_connections(
            connections=self.package.config.database.connections,
            default=self.package.config.database.default
        )

        # Define all tables/models used by this package.
        # The goal is to load up all SQLAlchemy tables for complete metedata definitions.
        # If you separate tables vs models use self.tables(['myapp.database.tables.*])
        # If you use models only, or models with inline tables then use self.models(['myapp.models.*])
        # Optionally define your own models/__init__.py and import myapp.models to load every table.
        # Order does not matter as they are sorted topologically for ForeignKey dependencies
        # Using __init__.py now, so just import it
        #from uvicore.auth import models
        #dump(self.package)
        self.register_db_models(['uvicore.auth.models'])

        # Define data seeders
        # NO - Auth shouldn't do its own seeding.  Let the app do it all.
        # You think?  What if a package is an app, then it runs seeders, but if that app is used
        # inside another package, you can't stop it from seeding.  Need to figure out overrideing seeders array better
        self.register_db_seeders(['uvicore.auth.database.seeders.seed'])

        # Define view and asset paths and configure the templating system
        self.define_views()

        # Define Web and API routes and prefixes
        self.define_routes()

    def define_views(self) -> None:
        """Define view and asset paths and configure the templating system"""

        # Define view paths
        #self.views(['uvicore.auth.http.views'])

        # Define public paths
        self.register_http_public(['uvicore.auth.http.public'])

        # Define asset paths
        self.register_http_assets(['uvicore.auth.http.public.assets'])

    def define_routes(self) -> None:
        """Define Web and API routes and prefixes"""

        # Define web routes
        # self.register_http_web_routes(
        #     module='uvicore.auth.http.routes.web.Web',
        #     prefix=self.package.config.web.prefix,
        #     #name_prefix=None,
        # )

        # Define api routes
        self.register_http_api_routes(
            module='uvicore.auth.http.routes.api.Api',
            prefix=self.package.config.api.prefix,
            #name_prefix='api',
        )



import uvicore
from uvicore.auth.user_info import UserInfo
from uvicore.support.hash import sha1
from uvicore.contracts import UserProvider
from uvicore.support.dumper import dump, dd
from uvicore.auth.support import password as pwd
from uvicore.http.request import HTTPConnection
from uvicore.typing import List, Union, Any, Dict
from uvicore.auth.models.user import User as UserModel
from uvicore.auth.models.group import Group
from datetime import datetime


@uvicore.service()
class Orm(UserProvider):
    """Retrieve and validate user from uvicore.auth ORM User model during Authentication middleware

    This is NOT a stateless user provider as it queries the user, groups, roles tables from a database.
    """

    # def __init__(self):
    #     # Only need for an __init__ override is to modify field mappings
    #     super().__init__()

    #     # Temp, until I add username to ORM model
    #     self.field_map['username'] = 'email'

    async def _retrieve_user(self,
        key_name: str,
        key_value: Any,
        request: HTTPConnection,
        *,
        password: str = None,

        # Parameters from auth config
        anonymous: bool = False,
        includes: List = None,

        # Must have kwargs for infinite allowed optional params, even if not used.
        **kwargs,
    ) -> UserInfo:
        """Retrieve user from backend"""

        # Cache store
        # Array store is much faster than redis and since this is called at the
        # middleware level on every request, we want it to be as performant as possible.
        # Set to None to use config default.
        cache_store = 'array'

        # Get password hash for cache key.  Password is still required to pull the right cache key
        # or else someone could login with an invalid password for the duration of the cache
        password_hash = '/' + sha1(password) if password is not None else ''

        # Check if user already validated in cache, if so, skip DB hits!
        # Don't do a cache.has() becuase cache.get() already does it
        # and because of the array store _expire() it's a bit expensive.
        # Eliminating the duplicate has() saved me about 1500 req/sec on wrk benchmark
        cache_key = 'auth/user/' + str(key_value) + password_hash
        user = await uvicore.cache.store(cache_store).get(cache_key)
        if user:
            # User is already validated and cached
            # Retrieve user from cache, no password check required because cache key has password has in it
            #dump('Cached authentication middleware User found, load from cache!')
            return user

        # Interesting.  With a heavy 'wrk' performance test you can see this hit
        # a dozen times on the first run.  Because of await, while caching is
        # trying to take place, many other request are comming in and processing.
        # We actually hit the DB a dozen times before the first request is cached.
        # This is why we do 'wrk' at least twice, to "warm up" the cache.  Also
        # with 'array` cache store and gunicorn, its actually one cache registery
        # per THREAD unlinke the shared redis backend which is just one cache.  So
        # with gunicorn and array caching you will see at least N cache misses.
        # But even still you see several more due to the concurrency of wrk and the
        # time it takes for await to set the cache.
        # dump('UNcached authentication middleware User, load from DB')

        # ORM is currently thworing a Warning: Truncated incorrect DOUBLE value: '='
        # when using actual bool as bit value.  So I convert to '1' or '0' strings instead
        disabled = '1' if anonymous else '0'

        # Cache not found.  Query user, validate password and convert to user class
        find_kwargs = {key_name: key_value}
        db_user = await (UserModel.query()
            .include(*includes)
            .where('disabled', disabled)
            #.show_writeonly(['password'])
            .show_writeonly(True)
            .find(**find_kwargs)
        )

        # User not found or disabled.  Return None means not verified or found.
        if not db_user: return None

        # If we are checking passwords and the db_user has NO password, user cannot be logged into
        if password is not None and db_user.password is None: return None

        # If password, validate credentials
        if password is not None:
            if not pwd.verify(password, db_user.password):
                # Invalid password.  Return None means not verified or found.
                return None

        # Get users groups->roles->permissions (roles linked to a group)
        groups = []
        roles = []
        permissions = []
        if 'groups' in includes:
            user_groups = db_user.groups
            if user_groups:
                for group in user_groups:
                    groups.append(group.name)
                    if not group.roles: continue
                    for role in group.roles:
                        roles.append(role.name)
                        if not role.permissions: continue
                        for permission in role.permissions:
                            permissions.append(permission.name)

        # Get users roles->permissions (roles linked directly to the user)
        if 'roles' in includes:
            user_roles = db_user.roles
            if user_roles:
                for role in user_roles:
                    roles.append(role.name)
                    if not role.permissions: continue
                    for permission in role.permissions:
                        permissions.append(permission.name)

        # Unique groups, roles and permissions (sets are unique)
        groups = sorted(list(set(groups)))
        roles = sorted(list(set(roles)))
        permissions = sorted(list(set(permissions)))

        # Set super admin, existence of 'admin' permission
        # Fixme, there is a 'superadmin' field on the roles table.
        # If user is in any role with superadmin=True they are a superadmin
        superadmin = False
        if 'admin' in permissions:
            # No need for any permissinos besides ['admin']
            permissions = ['admin']
            superadmin = True

        # Build UserInfo dataclass with REQUIRED fields
        user = UserInfo(
            id=db_user.id or '',
            uuid=db_user.uuid or '',
            username=db_user.username or '',
            email=db_user.email or '',
            first_name=db_user.first_name or '',
            last_name=db_user.last_name or '',
            title=db_user.title or '',
            avatar=db_user.avatar_url or '',
            groups=groups or [],
            roles=roles or [],
            permissions=permissions or [],
            superadmin=superadmin or False,
            authenticated=not anonymous,
        )

        # Save to cache
        if anonymous and cache_store == 'array':
            # If anonymous user, set cache to NEVER expire. Why?  Because
            # the anonymouse user will never change, no need to have it expire from cache.
            # This only works if cache store is 'array' because it will die when we kill
            # the program.  If cache store is 'redis', we want it to expire in case anyone
            # changes what the anonymous user in the DB looks like at some point.
            await uvicore.cache.store(cache_store).put(cache_key, user, seconds=0)
        else:
            # User is a valid user, cache it using configs default TTL expire
            # Or user is anonymous but cache_store is redis, we need to expire in redis.
            await uvicore.cache.store(cache_store).put(cache_key, user)

        # Return to user
        return user

    async def create_user(self, request: HTTPConnection, **kwargs):
        """Create new user in backend"""

        # Consistent kwargs come from our JWT auto_create_user_jwt_mapping config
        # uuid, username, email, first_name, last_name
        # title, avatar, creator_id, groups (array)

        # Pop groups from kwargs
        groups = kwargs.pop('groups')

        # Set other kwargs values
        kwargs['disabled'] = False
        kwargs['login_at'] = datetime.now()

        # Translate avatar
        kwargs['avatar_url'] = kwargs.pop('avatar')

        # Build user model
        user = UserModel(**kwargs)

        # Get actual groups in backend from groups array
        real_groups = await Group.query().where('name', 'in', groups).get()

        # Save user
        await user.save()

        # Link real_groups
        await user.link('groups', real_groups)

        # Return new backend user (not actual Auth user class)
        return user

    async def sync_user(self, request: HTTPConnection, **kwargs):
        """Sync user and group linkage to backend"""
        # Sync not only name changes, but also group linkage

        # Consistent kwargs come from our JWT auto_create_user_jwt_mapping config
        # uuid, username, email, first_name, last_name
        # title, avatar, creator_id, groups (array)

        # Get username
        username = kwargs['username']

        # Get actual backend user
        user = await UserModel.query().show_writeonly(['password']).include('groups').find(username=username)

        # If we have successfully logged in, we are not disabled
        user.disabled = False
        user.login_at = datetime.now()

        # Pop groups from kwargs
        groups = kwargs.pop('groups')

        # Remove other kwargs items
        del kwargs['creator_id']

        # Translate avatar
        kwargs['avatar_url'] = kwargs.pop('avatar')

        # Add all other kwargs to user
        for key, value in kwargs.items():
            setattr(user, key, value)

        # Save user record
        await user.save()

        # UNLINK any groups from DB that are not listed in groups JWT array
        if groups and user.groups:
            unlink_groups = []
            for g in user.groups:
                if g.name not in groups:
                    unlink_groups.append(g)
            if unlink_groups: await user.unlink('groups', unlink_groups)

        # LINK any groups in JWT that are not linked in DB
        if groups:
            # Get actual groups in backend from groups array
            real_groups = await Group.query().where('name', 'in', groups).get()
            real_group_names = [x.name for x in real_groups]

            # Convert users existing groups to list of names
            user_group_names = []
            if user.groups:
                user_group_names = [x.name for x in user.groups]

            link_groups = []
            for g in real_groups:
                if g.name not in user_group_names:
                    link_groups.append(g)
            if link_groups: await user.link('groups', link_groups)

        # Return new backend user (not actual Auth user class)
        return user



from uvicore.typing import Dict

class Connection(Dict):
    """Database Connection Definition"""

    # These class level properties for for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    name: str
    backend: str
    driver: str
    dialect: str
    host: str
    port: int
    database: str
    username: str
    password: str
    prefix: str
    metakey: str
    url: str
    is_async: bool



#from abc import ABC
#from dataclasses import dataclass
# @dataclass
# class Connection(ABC):
#     name: str
#     #default: bool
#     driver: str
#     dialect: str
#     host: str
#     port: int
#     database: str
#     username: str
#     password: str
#     prefix: str
#     metakey: str
#     url: str




from uvicore.typing import Dict, List

class Email(Dict):
    """Email Message Definition"""

    # These class level properties for for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    to: List
    cc: List
    bcc: List
    from_name: str
    from_address: str
    subject: str
    html: str
    text: str
    attachments: List



import uvicore
import inspect
import importlib
from uvicore.support import module
from uvicore.container import Binding
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Ioc as IocInterface
from uvicore.typing import Any, Callable, List, Optional, Type, TypeVar, Dict, Union

T = TypeVar('T')


class Ioc(IocInterface):
    """Inversion of Control private class.

    Do not import from this location.
    Use the uvicore.ioc singleton global instead."""

    @property
    def bindings(self) -> Dict[str, Binding]:
        return self._bindings

    @property
    def overrides(self) -> Dict[str, str]:
        # Merge app config bindings with registered overrides (app config wins)
        app_config_overrides = self._app_config.dotget('overrides.ioc_bindings') or {}
        overrides = {**self._overrides, **app_config_overrides}
        return overrides

    @property
    def aliases(self) -> Dict[str, str]:
        return self._aliases

    def __init__(self, app_config: Dict) -> None:
        self._bindings: Dict[str, Binding] = Dict()
        self._aliases: Dict[str, str] = Dict()
        self._app_config = app_config
        self._overrides: Dict[str, str] = Dict()


        # Add default binding specific to uvicore framework
        # Only some early defaults are here.  The rest are bound in
        # their service providers register() method
        # NO - Deprecated now that I can bind default and make
        self.bind_map({
            #'Application': {
            #    'object': 'uvicore.foundation.application._Application',
            #    'singleton': True,
            #    'aliases': ['App', 'app', 'application'],
            #},
            #'ServiceProvider': {
            #    'object': 'uvicore.package.provider._ServiceProvider',
            #    'aliases': ['service', 'provider'],
            #},
            #'Package': {
            #    'object': 'uvicore.package.package._Package',
            #    'aliases': ['package'],
            #},
            #'Dispatcher': {
            #    'object': 'uvicore.events.dispatcher._Dispatcher',
            #    'singleton': True,
            #    'aliases': ['dispatcher', 'Event', 'event', 'Events', 'events'],
            #},
        })

    #def config(self, config: Dict) -> None:
    #    self._app_config = config

    def binding(self, name: str = None, *, type: str = None, include_overrides: bool = True) -> Union[Binding, Dict]:
        if name:
            # Get one binding by name
            if name in self.bindings:
                return self.bindings[name]
            elif name in self.aliases:
                return self.bindings[self.aliases[name]]
        elif type:
            # Get all binding of the specified type
            #return [binding for binding in self.bindings.values() if binding.type.lower() == type.lower()]
            bindings = Dict({key:binding for key, binding in self.bindings.items() if binding.type.lower() == type.lower()})
            if include_overrides:
                return bindings

            # Strip out overridden bindings
            new_bindings = Dict()
            for key, binding in bindings.items():
                # Ignore BASE override models
                if binding.path != key: continue

                new_bindings[key] = binding
            return new_bindings


    def make(self, name: str, default: Callable[[], T] = None, **kwargs) -> T:
        if default is not None and self.binding(name) is None:
            # Default was provided and no binding currently exists
            # Bind the default provided but look for bindings override in app_config
            #object = default
            #bindings = self._app_config.get('bindings') or {}
            #object = bindings.get(name) or default
            object = self.overrides.get(name) or default
            self.bind(name, object, **kwargs)

        binding = self.binding(name)
        if not binding:
            # No binding set yet.  If we simply try to import the file only, it may have a
            # decorator that will bind itself.  If no binding found even after import, treat as not found
            if '.' in name: module.load(name)

            # Check binding again
            binding = self.binding(name)
            if not binding:
                raise ModuleNotFoundError("Could not find IoC name '{}' in mapping.".format(name))

        # If object is not defined, dynamically import it on first make (deferred)
        if not binding.object:
            # If object is None, dynamically import object from path
            binding.object = module.load(binding.path).object

        # Determine type
        is_class = inspect.isclass(binding.object)
        is_singleton = is_class and binding.singleton
        kwargs = binding.kwargs or {}

        # Odd case
        if not is_class and binding.singleton and hasattr(binding.object, '__class__') and '.' in str(getattr(binding.object, '__class__')):
            # If you override a singlton with another singleton (in the case of overriding a table for example)
            # You get an odd case where the binding object is the singleton itself.  So here we detect if the object
            # should be a singleton, and is NOT a class (because its already a singleton instance) and the object
            # is an instance (meaning it has a __class__ attribute) then we need to swap the instance with the object
            # and set the objects name to the instances __class__
            # By checking if __class__ has a . in it we skip over if someone accidentally added a singleton to a function or method
            # In case you are wondering, the singleton of the original IS the same singleton as the override!
            binding.instance = binding.object
            binding.object = binding.instance.__class__
            is_singleton = True

        # Instantiate a singleton only once
        made = None
        if is_singleton:
            if not binding.instance:
                if binding.factory:
                    factory = module.load(binding.factory).object
                    binding.instance = factory().make(binding.object, **kwargs)
                else:
                    binding.instance = binding.object(**kwargs)
            made = binding.instance

        # Instantiate a non-singleton every time
        # Unless there is no factory and no kwargs, simply return the object class
        elif is_class:
            if binding.factory:
                if type(binding.factory) == str:
                    # String factory, dynamically import it
                    factory = module.load(binding.factory).object
                else:
                    # Direct class factory
                    factory = binding.factory
                made = factory().make(binding.object, **kwargs)
            elif binding.kwargs:
                made = binding.object(**kwargs)
            else:
                made = binding.object

        # Bind is not a class.  Must be a method or module, return it
        else:
            made = binding.object

        # Return made object
        return made

    def bind_from_decorator(self, cls, name: str = None, *, object_type: str = None, factory: Any = None, kwargs: Dict = None, singleton: bool = False, aliases: List = []) -> None:
        """Bind from a decorator"""
        # Check for an override binding in the running app_config
        # These can also be defined in a service provider self.binding_override() method
        override = self.overrides.get(name)

        # If override and they aren't the same class
        if override and override != name:
            # Override OBJECT path defined in config or service provider self.binding_override() method
            self.bind(name=name, object=override, object_type=object_type, override=True, factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)

            # Also bind the original so I can import it to override it.  Originals should never be a singleton
            # We solve circular dependencies by adding the cls right to the binding, so it never has to import it!
            self.bind(name + '_BASE', cls, object_type=object_type, override=True, factory=factory, kwargs=kwargs, singleton=False, aliases=aliases)
        else:
            # No override.  Check if binding already exists.
            existing = self.binding(name)
            if existing and existing.path == name and existing.object is None:
                # Binding already exists and decorators never override existing bindings.
                # If existing binding is the same class as this decorator is on add in the
                # cls so .make() doesn't have to "import" the same class (causing circular
                # import issues).  If its not the same class, .make() will make and import it
                # as usual.  Use original decorators object_type and singeton because those
                # should NEVER be different (user error might override wrong).  If user provides no
                # aliases, use decorators.
                existing.object = cls
                existing.type = object_type
                existing.singleton = singleton
                if not existing.aliases: existing.aliases = aliases
                if not existing.kwargs: existing.kwargs = kwargs
                if not existing.factory: existing.factory = factory
            else:
                # Binding does not already exist, create it from decorator
                self.bind(name, cls, object_type=object_type, override=False, factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)

            #self.bind(name, cls, object_type=object_type, override=False, factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)
            #return cls

        # else:
        #     # Binding already exists and decorators never override existing bindings.  If existing binding is the same class as this decorator is on
        #     # add in the cls so .make() doesn't have to "import" the same class (causing circular import issues).  If its not the same class, .make() will
        #     # make and import it as usual.
        #     if self.bindings[name].path == name and self.bindings[name].object is None:
        #         self._bindings[name].object = cls
        #         self._bindings[name].type = object_type

        # Finally return the actual bind make, which if overridden, could be a completely different object!
        return self.make(name)

    def bind(self, name: str = None, object: Any = None, *, object_type: str = 'service', override: bool = True, factory: Any = None, kwargs: Dict = None, singleton: bool = False, aliases: List = []) -> None:
        # Decorator Usage
        if object is None:
            def decorator(cls):
                bind_name = name or cls.__module__ + '.' + cls.__name__
                return self.bind_from_decorator(cls, name=bind_name, object_type=object_type, factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)
            return decorator

        # Add each aliases to list of all aliases
        for alias in aliases:
            self._aliases[alias] = name

        # Set path and object based on str or actual class
        path = None
        if type(object) == str:
            path = object
            object = None

        if path is None and object is not None:
            if hasattr(object, '__module__') and hasattr(object, '__name__'):
                path = object.__module__ + '.' + object.__name__
            else:
                path = name

        # Add binding, obeying override
        if override == True or name not in self.bindings:
            #print(name, '----', path)
            self._bindings[name] = Binding(
                path=path,
                object=object,
                instance=None,
                type=object_type,
                factory=factory,
                kwargs=kwargs,
                singleton=singleton,
                aliases=aliases,
            )

    def bind_override(self, name: str, object: str):
        """Add a binding override to an array to check later"""
        self._overrides[name] = object

    def bind_map(self, mapping: Dict[str, Dict]) -> None:
        # bind_map is not used anymore, though could be cool if passed through from provider class as well, if ever
        for name, options in mapping.items():
            self.bind(name, **options)

    def alias(self, src: str, dest: str) -> None:
        if dest not in self.bindings:
            raise Exception('Could not find IoC binding '.format(dest))
        if src not in self.bindings[dest]:
            self.bindings[dest].aliases.append(src)



from abc import ABC, abstractmethod
from uvicore.contracts.user_info import UserInfo
from uvicore.typing import Optional, Tuple, Dict, List, Union
try:
    from starlette.requests import HTTPConnection
except ImportError:  # pragma: nocover
    HTTPConnection = None  # type: ignore


class Authenticator(ABC):

    @abstractmethod
    async def authenticate(self, conn: HTTPConnection) -> Union[UserInfo, bool]:
        pass

    @abstractmethod
    async def retrieve_user(self, username: str, password: str, provider: Dict) -> Optional[UserInfo]:
        """Retrieve user from User Provider backend"""
        pass

    @abstractmethod
    async def create_user(self, provider: Dict, request: HTTPConnection, **kwargs):
        """Create new user in backend"""
        pass

    @abstractmethod
    async def sync_user(self, provider: Dict, request: HTTPConnection, **kwargs):
        """Sync user and group linkage to backend"""
        pass

    @abstractmethod
    def auth_header(self, request) -> Tuple[str, str, str]:
        """Extract authorization header parts"""
        pass



# Do not import future here or http/bootstrap.py get_type_hints fails
# See https://bugs.python.org/issue41249
# NO from __future__ import annotations

import json
import uvicore
from uvicore.http.request import Request
from uvicore.http.params import Query
from uvicore.typing import Optional, Union, List, Tuple, Generic, TypeVar, Dict
from uvicore.orm.query import OrmQueryBuilder
from uvicore.contracts import AutoApi as AutoApiInterface
from uvicore.support.dumper import dump, dd
from uvicore.http.exceptions import PermissionDenied
from uvicore.contracts import UserInfo
from uvicore.http.exceptions import BadParameter

E = TypeVar("E")

# Default page size and maximum from app config
page_size_default = uvicore.config.app.api.page_size or 25
page_size_max = uvicore.config.app.api.page_size_max or 100


@uvicore.service()
class AutoApi(Generic[E], AutoApiInterface[E]):
    def __init__(self,
        Model: E,
        scopes: List = [],
        *,
        request: Request,
        include: Optional[List[str]] = None,
        find: Optional[str] = None,
        where: Optional[str] = None,
        or_where: Optional[str] = None,
        filter: Optional[str] = None,
        or_filter: Optional[str] = None,
        order_by: Optional[str] = None,
        sort: Optional[str] = None,
        page: Optional[int] = 1,
        page_size: Optional[int] = page_size_default,
    ):
        self.Model = Model
        self.scopes = scopes
        self.request = request
        self.user: UserInfo = request.user
        self.includes = self._build_include(include)
        self.find = self._build_find(find)
        self.page = page
        self.page_size = page_size
        if self.page_size > page_size_max: self.page_size = page_size_max

        # Where and filter JSON look identical, as does the ORM, all considered "whereables"
        self.wheres = self._build_whereable(where)
        self.or_wheres = self._build_whereable(or_where)
        self.filters = self._build_whereable(filter)
        self.or_filters = self._build_whereable(or_filter)

        # Order and sort JSON look identical, as does the ORM, all considered "sortables"
        self.order_bys = self._build_sortable(order_by)
        self.sorts = self._build_sortable(sort)

    @classmethod
    def findsig(
        request: Request,
        id: Union[str, int],
        include: Optional[List[str]] = Query([]),
        filter: Optional[str] = '',
        or_filter: Optional[str] = '',
    ):
        """AutoApi Find Function Signature"""
        pass

    @classmethod
    def getsig(
        request: Request,
        include: Optional[List[str]] = Query([]),
        find: Optional[str] = '',
        where: Optional[str] = '',
        or_where: Optional[str] = '',
        filter: Optional[str] = '',
        or_filter: Optional[str] = '',
        order_by: Optional[str] = '',
        sort: Optional[str] = '',
        page: Optional[int] = 1,
        page_size: Optional[int] = page_size_default,
    ):
        """AutoApi Get Function Signature"""
        pass


    def orm_query(self) -> OrmQueryBuilder[OrmQueryBuilder, E]:
        """Start a new Uvicore ORM Model QueryBuilder Query"""

        # Start empty query builder
        query = self.Model.query()

        # Include
        if self.includes: query.include(*self.includes)

        # Where
        if self.wheres: query.where(self.wheres)

        # OR Where
        if self.or_wheres: query.or_where(self.or_wheres)

        # Filter (children)
        if self.filters: query.filter(self.filters)

        # OR Filter (children)
        if self.or_filters: query.or_filter(self.or_filters)

        # Order by
        if self.order_bys: query.order_by(self.order_bys)

        # Sort (children)
        if self.sorts: query.sort(self.sorts)

        # Page and Page Size (ORM limit and offset)
        query.limit(self.page_size).offset(self.page_size * (self.page - 1))

        #query.key_by('id')

        # Return unfinished, still chainable query builder
        return query

    def guard_relations(self):
        # No includes, skip
        if not self.includes: return self

        # User is superadmin, allow
        if self.user.superadmin: return self

        # Loop each include and parts
        for include in self.includes:

            # Includes are split into dotnotation "parts".  We have to walk down each parts relations
            parts = [include]
            if '.' in include: parts = include.split('.')

            entity = self.Model
            for part in parts:

                # Get field for this "include part".  Remember entity here is not only Model, but a walkdown
                # based on each part (separated by dotnotation)
                field = entity.modelfields.get(part)  # Don't use modelfield() as it throws exception

                # Field not found, include string was a typo
                if not field: continue

                # Field has a column entry, which means its NOT a relation
                if field.column: continue

                # Field is not an actual relation
                if not field.relation: continue

                # Get actual relation object
                relation = field.relation.fill(field)

                # Get model permission string for this relation
                model_permissions = [relation.entity.tablename + '.read'] if self.scopes['overridden'] == False else self.scopes['read']
                #dump('CHILD MODEL PERMISSIONS:', model_permissions)

                # User must have ALL scopes defined on the route (an AND statement)
                authorized = True
                missing_scopes = []
                for scope in model_permissions:
                    if scope not in self.user.permissions:
                        authorized = False
                        missing_scopes.append(scope)

                # # Check if user has permissino to child relatoin (or superadmin)
                # if model_permission not in user.permissions:
                #     # I convert model_permissins to a list for consistency with other permission denied errors
                #     # although there will always be just one model_permission in this function
                #     # raise HTTPException(
                #     #     status_code=401,
                #     #     detail="Permission denied to {}".format(str([model_permission]))
                #     # )
                #     print('be')
                #     raise PermissionDenied(model_permission)

                if not authorized:
                    raise PermissionDenied(missing_scopes)

                # Walk down each "include parts" entities
                entity = relation.entity



            # if not field: continue
            # if not field.relation: continue
            # relation: Relation = field.relation.fill(field)
            # tablename = relation.entity.tablename
            # permission = tablename + '.read'

            # # Check if user has permissino to child relatoin (or superadmin)
            # if user.superadmin == False and permission not in user.permissions:
            #     raise HTTPException(
            #         status_code=401,
            #         detail="Access denied to {}".format(tablename)
            #     )

        # Chainable
        return self

    def _build_include(self, includes: List):
        if not includes: return
        results = []
        for include in includes:
            if ',' in include:
                results.extend(include.split(','))
            else:
                results.append(include)
        return results

    def _build_find(self, find_str: str) -> Dict:
        if not find_str: return None
        try:
            find = json.loads(find_str)
            if len(find) == 2:
                return {find[0]: find[1]}
        except Exception as e:
            raise BadParameter('Invalid find by named parameter, possibly invalid JSON?', exception=str(e), extra={'params': find_str})

    def _build_whereable(self, where_str: str) -> List[Tuple]:
        if not where_str: return None
        orm_wheres = []
        try:
            # Convert where string JSON to python object only if str, may already be a List from JSON payload not URL
            wheres = json.loads(where_str) if type(where_str) == str else where_str

            # Ensure we have a List[List]
            if type(wheres[0]) != list: wheres = [wheres]

            # Translate JSON object into ORM query builder
            for where in wheres:
                if len(where) == 2:
                    # No operator provided, just field and value, let ORM decide (ORM defaults to =)
                    orm_wheres.append((where[0], where[1]))
                elif len(where) == 3:
                    # Field, Operator and Value provided
                    orm_wheres.append((where[0], where[1], where[2]))

            return orm_wheres
        except Exception as e:
            raise BadParameter('Invalid where/or_where/filter/or_filter parameter, possibly invalid JSON?', exception=str(e), extra={'params': where_str})

    def _build_sortable(self, order_by_str: str) -> List[Tuple]:
        if not order_by_str: return None
        orm_order_bys = []
        try:
            # A simple order_by does not need to be json, could be ?order_by=id  If so, convert it to string [] list
            if order_by_str[0] != '[': order_by_str = '["' + order_by_str + '"]'

            # Convert order string JSON to python object only if str, may already be a List from JSON payload not URL
            order_bys = json.loads(order_by_str) if type(order_by_str) == str else order_by_str

            # Ensure we have a List[List]
            if type(order_bys[0]) != list: order_bys = [order_bys]

            # Translate JSON object into ORM query builder
            for order_by in order_bys:
                if len(order_by) == 1:
                    # No 'DESC' or 'ASC' defined, let ORM decide (ORM defaults to ASC)
                    orm_order_bys.append((order_by[0]))
                elif len(order_by) == 2:
                    # Both field and order (DESC, ASC) provided
                    orm_order_bys.append((order_by[0], order_by[1]))

            return orm_order_bys
        except Exception as e:
            raise BadParameter('Invalid order_by parameter, possibly invalid JSON?', exception=str(e), extra={'params': order_by_str})




from __future__ import annotations
import uvicore
from typing import Dict, List, Any, Optional
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Template as TemplateInterface

try:
    import jinja2
except ImportError:  # pragma: nocover
    jinja2 = None  # type: ignore

# This template system works for CLI templating and Web (starlette) templating
# But starlette is optional
try:
    from starlette.responses import Response
except ImportError:  # pragma: nocover
    class Response:
        pass


@uvicore.service('uvicore.templating.engine.Templates',
    aliases=['Templates', 'templates'],
    singleton=True,
)
class Templates(TemplateInterface):
    """Uvicore Template Engine"""

    @property
    def env(self) -> jinja2.Environment:
        return self._env

    @property
    def paths(self) -> List[str]:
        return self._paths

    @property
    def context_functions(self) -> Dict:
        return self._context_functions

    @property
    def context_filters(self) -> Dict:
        return self._context_filters

    @property
    def filters(self) -> Dict:
        return self._filters

    @property
    def tests(self) -> Dict:
        return self._tests

    def __init__(self) -> None:
        self._env = None
        self._paths = []
        self._context_functions = {}
        self._context_filters = {}
        self._filters = {}
        self._tests = {}

        # Only for starlette compatibility
        self.context_processors = {}

    def render(self, template_name: str, data: Optional[Dict] = {}) -> str:
        """Render a template as string (for CLI usage, not a Web response)"""
        template = self._env.get_template(template_name)
        return template.render(data)

    def render_web_response(self, name: str, context: dict, status_code: int = 200, headers: dict = None, media_type: str = None, background: BackgroundTask = None) -> _TemplateResponse:
        """Render a template as Web Response"""
        if "request" not in context:
            raise ValueError('context must include a "request" key')
        template = self._env.get_template(name)
        return _TemplateResponse(
            template,
            context,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
        )

    def _init(self) -> None:
        """Initialize the template engine (from bootstrap process)"""
        assert jinja2 is not None, "jinja2 is not installed"

        # Load our jinja2 environment
        loader = jinja2.FileSystemLoader(self.paths)
        self._env = jinja2.Environment(
            loader=loader,
            autoescape=True,
            keep_trailing_newline=True
        )

        # Register all user defined context processors
        self._register_context_processors()

    def _register_context_processors(self):
        """Register all user defined context processors"""
        # Add Context Functions
        for name, method in self.context_functions.items():
            #self._env.globals[name] = jinja2.contextfunction(method)
            self._env.globals[name] = jinja2.pass_context(method)
            #self._context_processors[name] = method

        # Add Context Filters
        for name, method in self.context_filters.items():
            #self._env.filters[name] = jinja2.contextfilter(method)
            self._env.filters[name] = jinja2.pass_context(method)
            #self._context_processors[name] = method

        # Add Filters
        for name, method in self.filters.items():
            self._env.filters[name] = method

        # Add Tests
        for name, method in self.tests.items():
            self._env.tests[name] = method

    def include_path(self, path) -> None:
        """Include template paths to search for"""
        if path not in self.paths:
            self._paths.append(path)

    def include_context_function(self, name: str, method: Any) -> None:
        """Include a context function"""
        self._context_functions[name] = method

    def include_context_filter(self, name: str, method: Any) -> None:
        """Include a context filter"""
        self._context_filters[name] = method

    def include_filter(self, name: str, method: Any) -> None:
        """Include a function"""
        self._filters[name] = method

    def include_test(self, name: str, method: Any) -> None:
        """Include a test"""
        self._tests[name] = method



class _TemplateResponse(Response):
    """Templated text/html Web Response"""

    media_type = "text/html"

    def __init__(
        self,
        template: typing.Any,
        context: dict,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ):
        self.template = template
        self.context = context
        content = template.render(context)
        super().__init__(content, status_code, headers, media_type, background)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.context.get("request", {})
        extensions = request.get("extensions", {})
        if "http.response.template" in extensions:
            await send(
                {
                    "type": "http.response.template",
                    "template": self.template,
                    "context": self.context,
                }
            )
        await super().__call__(scope, receive, send)



import uvicore
from uvicore.typing import Callable, Decorator


def model(name: str = None) -> Callable[[Decorator], Decorator]:
    def decorator(cls: Decorator) -> Decorator:
        # Bind this Model into the Ioc
        bind_name = name or cls.__module__ + '.' + cls.__name__
        new_cls = uvicore.ioc.bind_from_decorator(cls, name=bind_name, object_type='model', singleton=False)

        # Other things with a model, like add to array of models for auto-api perhaps?

        return new_cls
    return decorator



import uvicore
from uvicore.support.dumper import dump, dd
from dataclasses import dataclass
from uvicore.typing import Dict, Optional, Union, Callable


@uvicore.service()
class Event():

    # Defaults
    is_async: bool = False

    @classmethod
    @property
    def name(cls):
        name = str(cls).split("'")[1]
        # print(name, 'aaa')
        # if name == 'uvicore.events.event.Event':
        #     name = cls.__module__ + '.' + cls.__name__
        #print(name, 'xxx')
        return name

    @classmethod
    @property
    def description(cls):
        return cls.__doc__

    @classmethod
    def listen(cls, handler: Union[str, Callable], *, priority: int = 50):
        """Listen to to this event using this handler"""
        uvicore.events.listen(cls, handler, priority=priority)

    @classmethod
    def listener(cls, handler: Union[str, Callable], *, priority: int = 50):
        """Alias to Listen"""
        uvicore.events.listen(cls, handler, priority=priority)

    @classmethod
    def handle(cls, handler: Union[str, Callable], *, priority: int = 50):
        """Alias to Listen"""
        uvicore.events.listen(cls, handler, priority=priority)

    @classmethod
    def handler(cls, handler: Union[str, Callable], *, priority: int = 50):
        """Alias to Listen"""
        uvicore.events.listen(cls, handler, priority=priority)

    @classmethod
    def call(cls, handler: Union[str, Callable], *, priority: int = 50):
        """Alias to Listen"""
        uvicore.events.listen(cls, handler, priority=priority)

    def dispatch(self):
        """Fire off an event and run all listener callbacks"""
        uvicore.events._dispatch(self)

    async def dispatch_async(self):
        """Async fire off an event and run all async listener callbacks."""
        await uvicore.events._dispatch_async(self)

    async def codispatch(self):
        """Async fire off an event and run all async listener callbacks.  Alias for dispatch_async()."""
        await uvicore.events._dispatch_async(self)


    # No, logging should be a listener
    # def log(self, cls):
    #     uvicore.log.debug("Event " + str(cls.__class__.__module__) + " Dispatched")




#from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Generic, TypeVar, Union, List, Tuple, Any

class Mapper(ABC):

    @abstractmethod
    def column(self) -> str:
        """Convert a model field name into a table column name"""

    @abstractmethod
    def field(self):
        """Convert a table column name into a model field name"""

    @abstractmethod
    def model(self, perform_mapping: bool = True):
        """Convert a dict or List[dict] into a model or List[Model]

        Only maps table->model fields if perform_mapping = True, else assume already model fields.
        Works on a single record as dict - entity.mapper(DictModel).model()
        Works on a list of dict - entity.mapper(ListOfDictModel).model()
        Passes through if already a Model or List[Model]
        If mixed List of Dict and Model, converts all to Models
        """

    @abstractmethod
    def table(self):
        """Convert an model instance into a dictionary matching the tables columns (column mapper enabled)

        Works on a single model - model.mapper().table() or entity.mapper(model).table()
        Or on a list of model instances - entity.mapper(ListOfModelInstances).tabel()
        Or on a list of model dictionaries - entity.mapper(ListOfModelDict).table()
        Does not convert or recurse into relations
        """



import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.support.module import location, load
from uvicore.typing import List, Dict, OrderedDict, Union
from uvicore.contracts import Provider
from uvicore.support.collection import unique


class Templating(Provider):
    """Templating Provider Mixin"""

    def register_templating_paths(self, modules: List):
        # Default registration - template obeys view registration
        self.package.registers.defaults({'templates': True})

        # Register templates only if [templates] allowed
        if self.package.registers.templates:
            self.package.templating.paths = modules

    def register_templating_context_processors(self, items: Dict):
        # Default registration - template obeys view registration
        self.package.registers.defaults({'templates': True})

        # Register templates only if [templates] allowed
        if self.package.registers.templates:
            self.package.templating.context_processors = Dict(items)



"""
NOTES on how to get this working


Connection string has a new 'backend' set to 'sqlalchemy' by default

A connection for uvicore remote API might look like

# Example of ORM over Remote Uvicore API
'iam': {
    'backend': 'api',
    'driver': 'uvicore.orm.drivers.api.Api',
    'url': 'https://iam.example.com/api',
    'prefix': None
},

prefix required because the model will be user, but of backend prefix is auth_, url is auth_user


"""



import os
import uvicore
from uvicore.typing import Dict, Any
from uvicore.contracts import Connection
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Package as PackageInterface


@uvicore.service(aliases=['Package', 'package'])
class Package(PackageInterface):

    @property
    def config(self) -> Dict:
        """Helper to access this packages configs"""
        return uvicore.config.dotget(self.name)

    # def config(self, dotkey: str = None) -> Any:
    #     """Helper to access this packages configs"""
    #     if dotkey:
    #         return uvicore.config(self.name + '.' + dotkey)
    #     else:
    #         return uvicore.config(self.name)

    def connection(self, name: str = None) -> Connection:
        """Helper to access this packages connections"""
        try:
            if not self.database.connections: return None
            if name:
                return next(connection for connection in self.database.connections if connection.name == name)
            else:
                return next(connection for connection in self.database.connections if connection.name == self.database.connection_default)
        except:
            return None

    def folder_path(self, key: str) -> str:
        """Get a path for a package folder, accounting for package config paths overrides"""

        # Default paths
        defaults = Dict({
            #'base': base,
            'commands': 'commands',
            'config': 'config',
            'database': 'database',
            'migrations': 'database/migrations',
            'seeders': 'database/seeders',
            'tables': 'database/tables',
            'events': 'events',
            'http': 'http',
            'api': 'http/api',
            'assets': 'http/assets',
            'controllers': 'http/controllers',
            'routes': 'http/routes',
            'static': 'http/static',
            'views': 'http/views',
            'view_composers': 'http/composers',
            'jobs': 'jobs',
            'listeners': 'listeners',
            'models': 'models',
            'services': 'services',
            'support': 'support',
        })

        # Merge defaults with package paths
        paths = defaults.merge(self.config.paths)

        if key in paths:
            return os.path.realpath(self.path + '/' + paths[key])
        return None

        # Convert all to realpaths
        # for key, value in paths.items():
        #     paths[key] = os.path.realpath(self.path + '/' + value)




# # OBSOLETE
# @uvicore.service(aliases=['Package', 'package'])
# class Package(PackageInterface):

#     def config(self, dotkey: str = None):
#         if dotkey:
#             return uvicore.config(self.name + '.' + dotkey)
#         else:
#             return uvicore.config(self.name)

#     def connection(self, name: str = None):
#         if name:
#             return next(connection for connection in self.connections if connection.name == name)
#         else:
#             #return next(connection for connection in self.connections if connection.default == True)
#             if 'database' in self.config():
#                 return next(connection for connection in self.connections if connection.name == self.config('database.default'))



# IoC Class Instance
# No because not to be used by the public
#Package: _Package = uvicore.ioc.make('Package', _Package, aliases=['package'])

# Public API for import * and doc gens
#__all__ = ['_Package']



import uvicore
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd


@uvicore.provider()
class Logging(Provider):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Note about logger config
        # We cannot use the standard package config/logger.py here with proper
        # app config overrides because the logger is SUPER early in the bootstrapping
        # process.  I want the log available almost first thing, even in your packages
        # service provider register() and boot() methods.  This means if the logger config
        # were in this packages config/logger.py file you wouldn't be able to override it
        # as usual from your own app.  So instead, the logger config is in your actual app config

        # Register IoC bindings
        # self.bind('Logger', 'uvicore.logging.logger._Logger',
        #     aliases=['Log', 'log', 'logger'],
        #     kwargs={'config': self.app_config.get('logger')},
        #     singleton=True
        # )

        # Set uvicore.log global
        uvicore.log = uvicore.ioc.make('uvicore.logging.logger.Logger')

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        # Define service provider registration control
        # No - Never allow this packages registrations to be disabled from other configs
        # ?? course I don't have any registrations for now, maybe later



import uvicore
from uvicore.typing import Dict
from uvicore.events import Handler
from uvicore.support.module import load
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events.app import Booted as OnAppBooted


class Database(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the Database after the Application is Booted"""

        # Gather all connections, models, tables
        # No need to gather seeder classes, I get those in the ./uvicore db seed command
        connections = Dict()
        models = []; tables = []
        last_default = None; app_default = None
        for package in uvicore.app.packages.values():
            if not 'database' in package: continue

            # Get last defined default connection
            if package.database.connection_default: last_default = package.database.connection_default

            # Get running app default connection
            if package.main and package.database.connection_default: app_default = package.database.connection_default

            # Append connections
            connections.merge(package.database.connections)

            # Append models
            models.extend(package.database.models or [])

            # Append tables
            tables.extend(package.database.tables or [])

        # Initialize Database with all connections at once
        uvicore.db.init(app_default or last_default, connections)

        # Add all final merged connections to running_config
        uvicore.app.add_running_config('database.connections', uvicore.db.connections)

        # Dynamically Import models, tables and seeders
        for model in models: load(model)
        for table in tables: load(table)
